
import rrd
import settings
import tempfile
import subprocess
import os

"""
takes stats and convers to RRD
"""

DEBUG = False

class MongoRRD:

    IDENT = [ "customer" , "application" , "host" , "db" ]

    COUNTER_GROUPS = { "opcounters" : True , "btree" : True }
    COUNTER_NAMES = { "faults" : True }

    def __init__(self):
        self.rrd = rrd.RRD( settings.rrd_host )

    def processData( self , stats ):

        host = stats["host"]

        ident = {}
        for x in self.IDENT:
            if x in stats:
                ident[x] = stats[x]
        ident = self.genIdentString(ident)

        ######### serverStatus ############

        serverStatus = stats["serverStatus"]

        self.rrd.store( ident, "uptime" , serverStatus["uptime"] , )

        if "globalLock" in serverStatus:
            gl = serverStatus["globalLock"]
            if "currentQueue" in gl:
                self.rrd.storeAll( ident, gl["currentQueue"] , group="queue" )

        if "mem" in serverStatus:
            mem = serverStatus["mem"]
            if mem["supported"]:
                del mem["supported"]
                del mem["bits"]
                self.rrd.storeAll( ident , mem , group="memory" )


        for x in [ "connections" , "cursors" , "network" , "opcounters" , "asserts" ]:
            if x in serverStatus:
                self.rrd.storeAll( ident , serverStatus[x] , group=x  )

        self.rrd.storeAll( ident , serverStatus["indexCounters"]["btree"] , group="btree"  )

        if "extra_info" in serverStatus:
            ei = serverStatus["extra_info"]
            if "heap_usage_bytes" in ei:
                self.rrd.store( ident , "heap" , ei["heap_usage_bytes"] / ( 1024 * 1024 ) , group="memory"  )
            if "page_faults" in ei:
                self.rrd.store( ident , "faults" , ei["page_faults"] )

        ######### databases ############

        if "databases" in stats:
            m = stats["databases"]
            for dbname in m:
                myident = ident + "~~" + dbname
                mystats = m[dbname]

                print( "myident: " + myident )

                for simple in [ "collections" , "objects" , "indexes" , "avgObjSize" , "numExtents" ]:
                    self.rrd.store( myident , simple , mystats[simple] )

                for sz in [ "dataSize" , "storageSize" , "indexSize" , "fileSize" ]:
                    self.rrd.store( myident , sz , mystats[sz] , group="size" )


    def genIdentString(self,identDict,sep="~~"):
        if isinstance( identDict , basestring ) or isinstance( identDict , str ):
            return identDict
        s = ""
        for x in self.IDENT:
            if x in identDict:
                if len(s) > 0:
                    s += sep
                s += str(identDict[x])
        return s

    def genName(self,name,group,sep):


        if group and name:
            return group + sep + name
        elif group:
            return group
        elif name:
            return name

        raise Exception( "what?" )


    def graph(self,ident,group=None,name=None , outroot=None, timeType=rrd.TimeType.MINUTE):
        ident = self.genIdentString( ident  )

        # format 1: { "count" : [ ( <ts> , <num> ) , .... ] }
        isCounter = False;
        if group and group in self.COUNTER_GROUPS:
            isCounter = True
        if name and name in self.COUNTER_NAMES:
            isCounter = True
        raw = self.rrd.getDataFromQuery( ident, group=group,name=name, timeType=timeType , isCounter=isCounter )

        allDates = {}
        names = []

        data = {}
        for myname in raw:
            names.append( myname )

            new = {}
            for x in raw[myname]:
                # x is ( <ts> , <num> )
                ts = x[0]
                new[ts] = x[1]
                if ts in allDates:
                    allDates[ts] = allDates[ts] + 1
                else:
                    allDates[ts] = 1

            data[myname] = new

        #format 2 : { "count" : { "20101120" : 5 } }

        goodDates = []
        for x in allDates:
            if allDates[x] == len(names):
                goodDates.append( x )
        goodDates.sort()

        if len(goodDates) <= 1:
            return None

        ## write data file

        dataFile = tempfile.NamedTemporaryFile( prefix="mrrd" , suffix=".dat" , delete=not DEBUG)

        for date in goodDates:
            s = date + " "
            for myname in names:
                s += " " + str(data[myname][date])

            dataFile.write( s )
            dataFile.write( "\n" )

        dataFile.flush()

        ## write gnuplot script

        scriptFile = tempfile.NamedTemporaryFile( prefix="mrrd" , suffix=".plot" , delete=not DEBUG )

        scriptFile.write( "set term png enhanced small size 700,440\n" )
        scriptFile.write( "set autoscale \n" )

        # gen output name

        if not outroot:
            outroot = ""
        else:
            outroot += "/"

        shortDir = ident.replace( "~~" , "/" ) + "/"
        shortDir += timeType.name + "/"

        fulldir = outroot + shortDir
        if not os.path.exists( fulldir ):
            os.makedirs( fulldir )


        shortFileName = shortDir + self.genName(group,name,"." ) + ".png"
        longFileName = outroot + shortFileName

        if DEBUG:
            print( longFileName )

        scriptFile.write( "set output \"%s\" \n" % longFileName )

        # gen title
        title = ident.replace( "~~" , " " ) + " " + self.genName(group,name," ")
        if DEBUG:
            print( title )
        scriptFile.write( "set title \"%s\" \n" % title )

        # size
        scriptFile.write( "set size ratio 0.5\n" )

        # legend
        scriptFile.write( "set key below \n" )
        scriptFile.write( "set grid xtics ytics \n" )
        # x axis stuff
        scriptFile.write( "set xdata time \n" )
        scriptFile.write( "set timefmt '%Y%m%d-%H%M' \n" )
        scriptFile.write( "set format x '%d/%H:%M' \n" )
        scriptFile.write( "set xtics rotate \n" )

        # actual data
        scriptFile.write( "plot " )

        n = 2
        for x in names:
            if n > 2:
                scriptFile.write( " , " )
            scriptFile.write( " \"" + dataFile.name + "\" using 1:%d title '%s' with lines " % ( n , x ) )
            n += 1
        scriptFile.write( "\n" )
        scriptFile.write( "\n" )

        scriptFile.flush()

        ## run gnuplot

        ret = subprocess.call( ["/usr/bin/gnuplot" , scriptFile.name ] )
        if ret != 0:
            print( "error generating graph: " + longFileName )
            return None

        ## close/delete temp files
        scriptFile.close()
        dataFile.close()

        return shortFileName

    def graphAllForNode(self,i,outroot=None,timeType=None):
        i = self.genIdentString( i )
        types = self.rrd.getTypes( i )
        if not types:
            print( "no types for: " + str(i) )
            return []

        names = []

        for x in types:
            if timeType:
                if "group" in x:
                    names.append( self.graph( i , group=x["group"] , outroot=outroot, timeType=timeType) )
                else:
                    names.append( self.graph( i , name=x["name"] , outroot=outroot, timeType=timeType) )
            else:
                if "group" in x:
                    for tt in rrd.TimeType.ALL:
                        names.append( self.graph( i , group=x["group"] , outroot=outroot, timeType=tt) )
                else:
                    for tt in rrd.TimeType.ALL:
                        names.append( self.graph( i , name=x["name"] , outroot=outroot, timeType=tt ) )

        return names


    def graphAllNodes(self,outroot=None,timeType=None):
        names = []
        for i in self.rrd.getIdents():
            names += self.graphAllForNode(i,outroot=outroot,timeType=timeType)
        return names


if __name__ == "__main__":
    # this is for testing really

    m = MongoRRD()
    m.graphAllNodes("test")
